/*-*- mode:c;indent-tabs-mode:nil;c-basic-offset:2;tab-width:8;coding:utf-8 -*-│
│ vi: set et ft=c ts=2 sts=2 sw=2 fenc=utf-8                               :vi │
╞══════════════════════════════════════════════════════════════════════════════╡
│ Copyright 2020 Justine Alexandra Roberts Tunney                              │
│                                                                              │
│ Permission to use, copy, modify, and/or distribute this software for         │
│ any purpose with or without fee is hereby granted, provided that the         │
│ above copyright notice and this permission notice appear in all copies.      │
│                                                                              │
│ THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL                │
│ WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED                │
│ WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE             │
│ AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL         │
│ DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR        │
│ PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER               │
│ TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR             │
│ PERFORMANCE OF THIS SOFTWARE.                                                │
╚─────────────────────────────────────────────────────────────────────────────*/
#include "libc/intrin/bsr.h"
#include "libc/intrin/cxaatexit.h"
#include "libc/intrin/strace.h"
#include "libc/intrin/weaken.h"
#include "libc/macros.h"
#include "libc/mem/mem.h"
#include "libc/runtime/runtime.h"
#include "libc/sysv/errfuns.h"

__static_yoink("__cxa_finalize");

/**
 * Adds global destructor.
 *
 * Destructors are called in reverse order. They won't be called if the
 * program aborts or _exit() is called. Invocations of this function are
 * usually generated by the C++ compiler. Behavior is limitless if some
 * other module has linked calloc().
 *
 * @param fp is void(*)(T)
 * @param arg is passed to callback
 * @param pred can be non-null for things like dso modules
 * @return 0 on success or nonzero w/ errno
 * @note folks have forked libc in past just to unbloat atexit()
 */
int __cxa_atexit(void *fp, void *arg, void *pred) {
  /* asan runtime depends on this function */
  unsigned i;
  struct CxaAtexitBlock *b, *b2;
  __cxa_lock();
  b = __cxa_blocks.p;
  if (!b)
    b = __cxa_blocks.p = &__cxa_blocks.root;
  if (!~b->mask) {
    if (_weaken(calloc) &&
        (b2 = _weaken(calloc)(1, sizeof(struct CxaAtexitBlock)))) {
      b2->next = b;
      __cxa_blocks.p = b = b2;
    } else {
      __cxa_unlock();
      return enomem();
    }
  }
  i = bsr(~b->mask);
  b->mask |= 1u << i;
  b->p[i].fp = fp;
  b->p[i].arg = arg;
  b->p[i].pred = pred;
  __cxa_unlock();
  return 0;
}
